"""
Main function registry for Sharrock.
"""
from django.conf import settings
from django.template.defaultfilters import slugify
import inspect
import os.path

descriptor_registry = {}
resource_registry = {}

def get_module(module_name):
    """
    Imports and returns the named module.
    """
    module = __import__(module_name)
    components = module_name.split('.')
    for comp in components[1:]:
        module = getattr(module,comp)
    return module

def build_registry():
    """
    Builds the function descriptor registry.
    """
    for app_path in settings.INSTALLED_APPS:
        if app_path != 'sharrock': # don't load yourself
            try:
                module = get_module('%s.descriptors' % app_path)
                if is_package(module):
                    load_multiple_versions(app_path,module)
                else:
                    load_descriptors(app_path,module)
            except ImportError:
                pass # no descriptors in that module

def load_multiple_versions(app_path,package):
    """
    Loads multiple versions of an app's API.  Multiple versions are stored in submodules of
    a 'descriptors' package within the app.  (When there is only one version of an API,
    'descriptors' is a simple module).
    """
    for sublabel in package.__all__:
        submodule = get_module('%s.%s' % (package.__name__,sublabel))
        load_descriptors(app_path,submodule)

def load_descriptors(app_path,descriptor_module):
    """
    Loads descriptors in the module into the directory.
    """
    from sharrock.descriptors import Descriptor, Resource
    from sharrock.modelresource import ModelResource

    version = '0.1dev' # default version
    if hasattr(descriptor_module,'version'):
        version = getattr(descriptor_module,'version')
    
    module_deprecated = None
    if hasattr(descriptor_module,'deprecated'):
        # entire module is deprecated
        module_deprecated = descriptor_module.deprecated

    for name,attribute in inspect.getmembers(descriptor_module):
        if inspect.isclass(attribute) and issubclass(attribute,Descriptor) and not attribute is Descriptor:
            if not hasattr(attribute,'visible') or attribute.visible: # skip over descriptors with visible=False set
                descriptor_registry[(app_path,version,slugify(name))] = attribute(is_deprecated=module_deprecated) # put instance of the descriptor into the registry
        elif inspect.isclass(attribute) and issubclass(attribute,Resource) and not attribute is Resource and not attribute is ModelResource:
            descriptor_registry[(app_path,version,slugify(name))] = attribute(is_deprecated=module_deprecated) # put instance of resource into registry
    

def get_descriptor(app_label,version,descriptor_slug):
    """
    Gets the matching descriptor.
    """
    return descriptor_registry[(app_label,version,descriptor_slug)]

def is_package(module):
    """
    Checks if the specified module is a package.
    """
    return module.__file__.endswith('__init__.py') or module.__file__.endswith('__init__.pyc')

def directory(app_label=None,specified_version=None):
    """
    Creates a directory of service descriptors.
    """
    ensure_registry()
    
    from sharrock.descriptors import Resource
    d = {}
    for key, value in descriptor_registry.items():
        app,version,name = key
        if not app_label or app_label == app:
            app_dict = d.get(app,{})
            if not specified_version or specified_version == version:
                descriptors = app_dict.get(version,{'resources':[],'functions':[]})
                if issubclass(value.__class__,Resource):
                    descriptors['resources'].append(value)
                else:
                    descriptors['functions'].append(value)
                app_dict[version] = descriptors
            d[app] = app_dict
    
    return d

def resource_directory(app_label=None,specified_version=None):
    """
    Creates a directory for resources.
    """
    d = {}
    for key, value in resource_registry.items():
        app,version,name = key
        if not app_label or app_label == app:
            app_dict = d.get(app,{})
            if not specified_version or specified_version == version:
                resources = app_dict.get(version,[])
                resources.append(value)
                app_dict[version] = resources
            d[app] = app_dict
    
    return d

def ensure_registry():
    """
    Loads descriptors if the repository is unpopulated.
    """
    if not descriptor_registry:
        build_registry()




"""
URLs for Sharrock resource layer.
"""
from django.conf.urls.defaults import *

urlpatterns = patterns('sharrock.views',
    (r'^describe/(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<service_name>[\w-]+)\.(?P<extension>\w+)$','describe_service',{'service_type':'resource'}),
    (r'^describe/(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<service_name>[\w-]+)/$','describe_service',{'extension':'html','service_type':'resource'}),
    (r'^(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<resource_name>[\w-]+)/$','execute_resource',{'extension':'json'}),
    (r'^(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<resource_name>[\w-]+)\.(?P<extension>\w+)$','execute_resource'),
    (r'^(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<resource_name>[\w-]+)/list/$','execute_resource',{'extension':'json'}), # model resource url
    (r'^(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<resource_name>[\w-]+)/list\.(?P<extension>\w+)$','execute_resource'),  # model resource url
    (r'^(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<resource_name>[\w-]+)/create/$','execute_resource',{'extension':'json'}), # model resource url
    (r'^(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<resource_name>[\w-]+)/create\.(?P<extension>\w+)$','execute_resource'),  # model resource url
    (r'^(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<resource_name>[\w-]+)/(?P<model_id>\d+)/$','execute_resource',{'extension':'json'}), # model resource url
    (r'^(?P<app>[\w\.]+)/(?P<version>[\w\.]+)/(?P<resource_name>[\w-]+)/(?P<model_id>\d+)\.(?P<extension>\w+)$','execute_resource'), # model resource url
)


"""
Unit tests for Sharrock
"""
import unittest
from sharrock.client import HttpClient, ResourceClient, ModelResourceClient
from django.contrib.auth.models import User

class ClientTests(unittest.TestCase):
    """
    Client tests using example.
    """
    def setUp(self):
        """
        Runs before each test.
        """
        self.c = HttpClient('http://localhost:8000/api','sharrock_example','1.0')
    
    def tearDown(self):
        """
        Runs after each test.
        """
        self.c = None
    
    def test_hello_world(self):
        """
        Tests the hello world app.
        """
        result = self.c.helloworld()
        self.assertEquals(result,'Hello world!')

        new_result = self.c.helloworld(name='Loren')
        self.assertEquals(new_result,'Hello Loren!')
    
    def test_post_data(self):
        """
        Tests the post data service.
        """
        result = self.c.postdata(data={'foo':'bar'})
        self.assertEquals(result['grommit'],'bar')

class ResourceClientTests(unittest.TestCase):
    """
    Tests for resource client.
    """
    def setUp(self):
        """
        Runs before each test.
        """
        self.c = ResourceClient('http://localhost:8000/resources','sharrock_resource_example','1.0','meresource')
    
    def tearDown(self):
        """
        Runs after each test.
        """
        self.c = None
    
    def test_get(self):
        """
        Tests the GET op.
        """
        response = self.c.get()
        self.assertEquals('Get Method executed!',response)
    
    def test_post(self):
        """
        Tests the POST op.
        """
        response = self.c.post(params={'name':'posttest'})
        self.assertEquals('Posted posttest',response)
    
    def test_put(self):
        """
        Tests the PUT op.
        """
        response = self.c.put(params={'name':'puttest'})
        self.assertEquals('Put this:puttest',response)
    
    def test_delete(self):
        """
        Tests the delete op.
        """
        response = self.c.delete()
        self.assertEquals('Aaaarrrggghhhh!  I\'m meeelllltttiiinnnngggg!',response)

class ModelResourceClientTests(unittest.TestCase):
    """
    Tests a model resource.
    """
    def setUp(self):
        """
        Runs before each test.
        """
        self.c = ModelResourceClient('http://localhost:8000/resources','sharrock_modelresource_example','1.0','userresource')
        self.tom = User.objects.create(username='Tom')
        self.dick = User.objects.create(username='Dick')
        self.harry = User.objects.create(username='Harry')
    
    def tearDown(self):
        """
        Runs after each test.
        """
        self.harry.delete()
        self.dick.delete()
        self.tom.delete()
        self.c = None
    
    def test_list(self):
        """
        Tests the listing of a resource.
        """
        results = self.c.list()
        usernames = [result['username'] for result in results]
        self.assertTrue('Tom' in usernames)
        self.assertTrue('Dick' in usernames)
        self.assertTrue('Harry' in usernames)
    
    def test_get(self):
        """
        Tests the get function.
        """
        tom_dict = self.c.get(self.tom.pk)
        self.assertEquals(tom_dict['username'],self.tom.username)
    
    def test_create(self):
        """
        Tests create function.
        """
        zelda_dict = self.c.create(username='Zelda')
        zelda_model = User.objects.get(pk=zelda_dict['id'])
        self.assertEquals('Zelda',zelda_model.username)
    
    def test_update(self):
        """
        Tests the update function.
        """
        self.c.update(self.tom.pk,first_name='Thomas',last_name='Wayne')
        tom = User.objects.get(pk=self.tom.pk)
        self.assertEquals(tom.first_name,'Thomas')
        self.assertEquals(tom.last_name,'Wayne')
    
    def test_delete(self):
        """
        Tests the delete function.
        """
        tom_pk = self.tom.pk
        self.c.delete(tom_pk)
        self.assertRaises(User.DoesNotExist,User.objects.get,pk=tom_pk)




